from binascii import a2b_hex
from datetime import datetime
from typing import List, Optional, Union, get_args, get_origin

from .const import OBIS_NAMES, UNITS
from .sml import SmlTime


class CEndofSmlMsg:
    def __repr__(self):
        return '<EndofSmlMsg>'


EndofSmlMsg = CEndofSmlMsg()
NoneType = type(None)

INDENT = '    '


class SmlMsgBase:
    @classmethod
    def from_list(cls, _in: list):
        anno = cls.__annotations__
        c_req = len(anno)

        if not isinstance(_in, list) and len(_in) != c_req:
            raise ValueError(f'List with {c_req} entries expected, got {_in} ({type(_in)})')

        c = cls()
        for i, k in enumerate(cls.__annotations__.items()):
            k, _t = k

            # todo: add test that all Unions are short for Optional
            # Optional transforms to Union with two args, e.g. (str, None)
            origin = get_origin(_t)
            if origin is Union or origin is list:
                _t = get_args(_t)[0]

            v = _in[i]
            # if the passed value is a list it's either a new MsgField or
            # a list of MsgFields
            if isinstance(v, list):
                if origin is list:
                    for _i, _v in enumerate(v):
                        v[_i] = _t.from_list(_v)
                else:
                    v = _t.from_list(v)
            setattr(c, k, v)

        # todo: move this to testing
        i = 0
        for k, _t in anno.items():
            v = getattr(c, k)
            i += 1

            origin = get_origin(_t)
            args = get_args(_t)

            if origin is Union or origin is list:
                _t = args[0]

            if origin is list:
                for v in v:
                    assert isinstance(v, _t), f'{k} {v}: {type(v)} != {_t} for {cls}\n{_in}'
            else:
                if origin is Union and v is None:
                    continue
                if _t is SmlTime:
                    _t = (int, datetime)
                assert isinstance(v, _t), f'{k} {v}: {type(v)} != {_t} for {cls}\n{_in}'

        return c

    def format_msg(self, indent: int = 0):
        r = f'<{self.__class__.__name__}>\n'
        w = max(map(len, self.__dict__), default=0)
        for k, v in self.__dict__.items():
            if isinstance(v, SmlMsgBase):
                r += f'{INDENT * indent}{str(k):s} {v.format_msg(indent + 1)}'
            elif isinstance(v, list):
                r += f'{INDENT * indent}{str(k):s}: list\n'
                for e in v:
                    r += f'{INDENT * (indent + 1)}{e.format_msg(indent + 2)}'
            else:
                r += f'{INDENT * indent}{str(k):{w}s}: {v}\n'
        return r


class SmlMessage:
    transaction_id: str
    group_no: int
    abort_on_error: int
    message_body: SmlMsgBase
    crc16: int

    @classmethod
    def from_list(cls, _in: list) -> 'SmlMessage':
        assert isinstance(_in, list) and len(_in) == 6
        if not _in[5] is EndofSmlMsg:
            raise ValueError(f'Last entry is not {EndofSmlMsg}: {_in[5]}')
        # Load Msg Contents
        _in[3] = get_msg_cls(_in[3])
        return cls(*_in)

    def __init__(self, transaction_id: str, group_no: int, abort_on_error: int, message_body: List[SmlMsgBase],
                 crc16: int, eom: CEndofSmlMsg = None):
        self.transaction_id: str = transaction_id
        self.group_no: int = group_no
        self.abort_on_error: int = abort_on_error
        self.message_body: List[SmlMsgBase] = message_body
        self.crc16: int = crc16

    def format_msg(self):
        r = f'{self.__class__.__name__}\n'
        w = max(map(len, self.__dict__), default=0)
        for k, v in self.__dict__.items():
            if isinstance(v, SmlMsgBase):
                r += f'{INDENT}{str(k):s} {v.format_msg(2)}'
            else:
                r += f'{INDENT}{str(k):{w}s}: {v}\n'
        return r


class SmlOpenResponse(SmlMsgBase):
    codepage: Optional[str]
    client_id: Optional[str]
    req_file_id: str
    server_id: str
    ref_time: Optional[str]
    sml_version: Optional[int]


class SmlCloseResponse(SmlMsgBase):
    global_signature: Optional[str]


class SmlListEntry:

    @classmethod
    def from_list(cls, _in):
        if not isinstance(_in, list) and len(_in) != 7:
            raise ValueError(f'List with 7 entries expected, got {_in} ({type(_in)})')
        if _in[0] is None:
            raise ValueError('Obis field is required!')
        if _in[5] is None:
            raise ValueError('value is required!')

        # Maybe it's ascii so we try to decode it
        if isinstance(_in[5], str):
            v = a2b_hex(_in[5]).decode(errors='ignore')
            if v.isalnum():
                _in[5] = v

        return cls(*_in)

    def __init__(self, obis: str, status: Optional[int], val_time, unit: Optional[int], scaler: Optional[int] = None,
                 value: Union[None, str, int, float] = None, value_signature: Optional[str] = None):
        self.obis = obis
        self.status: Optional[int] = status
        self.val_time = val_time
        self.unit: Optional[int] = unit
        self.scaler: Optional[int] = scaler
        self.value = value
        self.value_signature = value_signature

    def __repr__(self):
        r = []
        for k, v in self.__dict__.items():
            if v is not None:
                r.append(f'{k}: {v}')
        return f'<{", ".join(r)}>'

    def format_msg(self, indent: int = 0):
        r = f'<{self.__class__.__name__}>\n'
        w = max(map(len, self.__dict__), default=0)
        for k, v in self.__dict__.items():
            r += f'{INDENT*indent}{str(k):{w}s}: {v}\n'

        summary = ''
        if self.unit:
            val = self.get_value()
            u = UNITS.get(self.unit)
            if u is None:
                u = f" ?:{self.unit}"
            summary += f'{val}{u}'

        desc = OBIS_NAMES.get(self.obis)
        if desc is not None:
            summary += f'{" " if summary else ""}({desc})'
        if summary:
            r += f'{INDENT*indent}-> {summary:s}\n'
        return r

    def get_value(self) -> float:
        if self.scaler is None:
            return self.value

        return round(self.value * 10**self.scaler, abs(self.scaler) + 3)


class SmlGetListResponse(SmlMsgBase):
    client_id: Optional[str]
    sever_id: str
    list_name: Optional[str]
    act_sensor_time: Optional[SmlTime]
    val_list: List[SmlListEntry]
    list_signature: Optional[str]
    act_gateway_time: Optional[SmlTime]


def get_msg_cls(_in: list) -> SmlMsgBase:
    types = {
        0x0100: 'SmlOpenRequest',
        0x0101: SmlOpenResponse,
        0x0200: 'SmlCloseRequest',
        0x0201: SmlCloseResponse,
        0x0300: 'SmlGetProfilePackRequest',
        0x0301: 'SmlGetProfilePackResponse',
        0x0400: 'SmlGetProfileListRequest',
        0x0401: 'SmlGetProfileListResponse',
        0x0500: 'SmlGetProcParameterRequest',
        0x0501: 'SmlGetProcParameterResponse',
        0x0600: 'SmlSetProcParameterRequest',
        0x0601: 'SmlSetProcParameterResponse',
        0x0700: 'SmlGetListRequest',
        0x0701: SmlGetListResponse,
        0xff01: 'SmlAttentionResponse',
    }

    if len(_in) != 2:
        raise ValueError(f'Not a class container: {_in}')

    _msg_type = _in[0]
    cls = types.get(_msg_type)
    if cls is None:
        raise ValueError(f'Unknown message type: {_msg_type:x}')

    if isinstance(cls, str):
        raise ValueError(f'{cls} is not yet supported!')

    return cls.from_list(_in[1])



from typing import Optional

from .crc import get_crc
from .sml_frame import SmlFrame


class CrcError(Exception):
    def __init__(self, msg: bytes, crc_msg: int, crc_calc: int):
        self.msg = msg
        self.crc_msg = crc_msg
        self.crc_calc = crc_calc

    def __repr__(self):
        return f'<{self.__class__.__name__} msg: {self.crc_msg:04x} calc: {self.crc_calc:04x}>'


class SmlStreamReader:
    MAX_SIZE = 50 * 1024

    def __init__(self):
        self.bytes: bytes = b''

    def add(self, _bytes: bytes):
        self.bytes += _bytes
        if len(self.bytes) > SmlStreamReader.MAX_SIZE:
            self.bytes = self.bytes[-1 * SmlStreamReader.MAX_SIZE:]

    def clear(self):
        self.bytes = b''

    def get_frame(self) -> Optional[SmlFrame]:
        start = self.bytes.find(b'\x1B\x1B\x1B\x1B\x01\x01\x01\x01')
        if start == -1:
            return None

        # if we start reading in the mid of a message
        if start != 0:
            self.bytes = self.bytes[start:]
            start = 0

        end = -1
        while (end := self.bytes.find(b'\x1B\x1B\x1B\x1B\x1A', end + 1)) != -1:
            pre = self.bytes[end - 4: end]
            if pre != b'\x1B\x1B\x1B\x1B':
                break

        if end == -1:
            return None

        end += 8
        if len(self.bytes) < end:
            return None

        # remove msg from buffer
        msg = self.bytes[start:end]
        self.bytes = self.bytes[end:]

        # Last three bytes are PADDING, CRC PART 1, CRC PART 2
        padding = msg[-3]

        # check crc
        crc_msg = msg[-2] << 8 | msg[-1]
        crc_calc = get_crc(msg[:-2])
        if crc_msg != crc_calc:
            raise CrcError(msg, crc_msg, crc_calc)

        return SmlFrame(msg[8: -1 * (8 + padding)].replace(b'\x1B\x1B\x1B\x1B\x1B\x1B\x1B\x1B', b'\x1B\x1B\x1B\x1B'))



from typing import List, Optional

from .sml_fields import EndofSmlMsg, SmlListEntry, SmlMessage


class InvalidBufferPos(Exception):
    pass


class SmlFrame:
    def __init__(self, buffer: bytes):
        self.bytes = buffer
        self.buffer = memoryview(buffer)
        self.buf_len = len(buffer)

        self.next_pos = 0

    def get_value(self, pos: Optional[int] = None):
        if pos is None:
            pos = self.next_pos

        # check start pos
        if pos >= self.buf_len:
            raise InvalidBufferPos(f'Start pos bigger than buffer: {pos} > {self.buf_len}')

        v = self.buffer[pos]

        # No value
        if v == 0x01:
            self.next_pos = pos + 1
            return None

        # Bool
        if v == 0x42:
            self.next_pos = pos + 2
            return bool(self.buffer[pos + 1])

        # End of a SmlMSg
        if v == 0x00:
            self.next_pos = pos + 1
            return EndofSmlMsg

        # types with dynamic size
        s_pos = pos
        is_long = bool(v & 0x80)
        _type = v & 0x70  # type
        _size = v & 0x0F  # size including the 1-byte tag
        while is_long:
            s_pos += 1
            v = self.buffer[s_pos]
            _size = _size << 4 | v & 0x0F
            is_long = bool(v & 0x80)

        # type is a list
        if _type == 0x70:
            self.next_pos = pos + 1
            return [None for i in range(_size)]

        # End position
        e_pos = pos + _size
        if e_pos > self.buf_len:
            raise InvalidBufferPos(f'Pos bigger than buffer: {e_pos} > {self.buf_len}')
        self.next_pos = e_pos

        # 0x50: signed integer, 0x60 unsigned integer
        if _type in (0x50, 0x60):
            return int.from_bytes(
                self.buffer[s_pos + 1:e_pos], byteorder='big', signed=True if _type == 0x50 else False
            )

        # 0x00: octet str
        if _type == 0x00:
            return self.buffer[s_pos + 1:e_pos].hex()

        raise ValueError('Unknown data type!')

    def parse_frame(self) -> List[SmlMessage]:
        ret = []
        while self.next_pos < self.buf_len:

            if not self.buffer[self.next_pos] == 0x76:
                raise ValueError(
                    f'No start of SML Message found at {self.next_pos}: 0x{self.buffer[self.next_pos]:x}\n'
                    f'{self.buffer.hex()}'
                )

            # This will always return a list
            val = self.get_value()
            self._parse_msg(val)
            if val[-1] != EndofSmlMsg:
                raise ValueError(f'Last Entry is not {EndofSmlMsg}! Something went wrong!')

            ret.append(SmlMessage.from_list(val))
        return ret

    def _parse_msg(self, parent_obj=None):
        # it's always a list now
        for i, _ in enumerate(parent_obj):
            parent_obj[i] = v = self.get_value()
            if isinstance(v, list):
                self._parse_msg(v)

    def get_obis(self) -> List[SmlListEntry]:
        """Returns all obis values in the frame without parsing the frame"""
        ret = []
        start = -1
        while (start := self.bytes.find(b'\x77\x07\x01', start + 1)) != -1:
            data = self.get_value(start)
            if not isinstance(data, list):
                continue

            self._parse_msg(data)
            ret.append(SmlListEntry.from_list(data))
        return ret



from datetime import datetime, timedelta, tzinfo
from typing import Optional


class SmlTzOffset(tzinfo):
    def __init__(self, utc_offset: int, dst_offset: int):
        self._utc_offset = timedelta(minutes=utc_offset)
        self._dst_offset = timedelta(minutes=dst_offset)

    def tzname(self, dt: Optional[datetime]) -> Optional[str]:
        return 'SmlTzOffset'

    def utcoffset(self, dt: Optional[datetime]) -> Optional[timedelta]:
        return self._utc_offset

    def dst(self, dt: Optional[datetime]) -> Optional[timedelta]:
        return self._dst_offset


class SmlTime:
    @classmethod
    def from_list(cls, _in):
        _t = _in[0]
        _v = _in[1]
        if _t == 1:
            return _v
        if _t == 2:
            return datetime.fromtimestamp(_v)
        if _t == 3:
            return datetime.fromtimestamp(_v[0], tz=SmlTzOffset(_v[1], _v[2]))
